/*
  uForth - A tiny ROMable 16-bit FORTH-like scripting language
          for microcontrollers.

  License for uForth 1.0 and later versions

  Copyright  2009 Todd Coram, todd@maplefish.com, USA.

  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:
  
  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <stdint.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "uforth.h"


#define DEBUG

struct iram iram;

/*
  Words must be under 64 characters in length
*/
#define WORD_LEN_BITS 0x3F  
#define IMMEDIATE_BIT (1<<7)
#define PRIM_BIT     (1<<6)

CELL uram[URAM_CELLS];

enum { /* Must be negative to prevent confusing with addresses  */
  LIT=-100, ABORT, DEF, IMMEDIATE, PICK, RPICK, SWAP, DUP, DROP, HERE, 
  IF, ELSE, THEN, DO, LOOP, LEAVE, I, J,
  ADD, SUB, MULT, DIV, AND, JMP, JMP_IF_ZERO, EXIT,
  OR, XOR, LSHIFT, RSHIFT, EQ_ZERO, COMPILE_FIND_WORD, FIND_WORD, EXEC,
  BEGIN, UNTIL, WHILE, REPEAT, EQ, LESS_THAN, GREATER_THAN,
  NOT, COMMA, RPUSH, RPOP, FETCH, STORE, COMMENT, COMMA_STRING,
  LESS_THAN_ZERO, CREATE, VARIABLE, CALLC };



/*
 Every entry in the dictionary consists of the following cells:
  [index of previous entry]  
  [flags, < 128 byte name byte count]
  [name [optional pad byte]... [ data ..]
*/
CELL find_word(char* s, bool *immediate, bool *prim);

void make_word(char *str, uint8_t str_len) {
  CELL my_head = dict->here;

  dict_append(dict->last_word_idx);
  dict_append(str_len);
  dict_append_string(str, str_len);
  dict_set_last_word(my_head);
#ifdef DEBUG_2
  printf("Making: %.*s : %d\n", str_len, str, dict->here);
#endif
}


void make_immediate(void) {
  dict_write((dict->last_word_idx+1), dict->d[dict->last_word_idx+1]|IMMEDIATE_BIT);
}


void store_prim(char* str, CELL val) {
  make_word(str,strlen(str));
  dict_append(val);
  dict_write((dict->last_word_idx+1), dict->d[dict->last_word_idx+1]|PRIM_BIT);
  /* Primitives don't need an exit.. we optimize away the call by inlining
     them.
  */
}


char next_char(void) {
  return *iram.inbufptr++;
}

char* uforth_next_word (void) {
  while (isspace(*iram.inbufptr)) ++iram.inbufptr;
  iram.currword = iram.inbufptr;
  while (*iram.inbufptr != 0 && !isspace(*iram.inbufptr)) ++iram.inbufptr;
  iram.currwordlen = iram.inbufptr - iram.currword;
  return iram.currword;
}

void uforth_abort(void) {
  if (iram.compiling) {
    dict_append(ABORT);
  }
  iram.compiling = 0;
  iram.didx = -1;
  iram.ridx = -1;
  iram.varidx = 1;	/* start at 1, since ram is pushed as negative */
  iram.ifnest = 0;
  iram.donest = 0;
  iram.leave_loop = 0;
}

void uforth_init(void) {
  uforth_abort();

#ifdef BOOTSTRAP
  /*
    Store our primitives into the dictionary.
  */
  store_prim("cf", CALLC);
  store_prim("immediate", IMMEDIATE);
  store_prim("abort", ABORT);
  store_prim("pick", PICK);
  store_prim("rpick", RPICK);
  store_prim("drop", DROP);
  store_prim("dup", DUP);
  store_prim("swap", SWAP);
  store_prim("0jmp?", JMP_IF_ZERO);
  store_prim("exec", EXEC);
  store_prim(",", COMMA);
  store_prim("here", HERE);
  store_prim("<0", LESS_THAN_ZERO);
  store_prim("+", ADD);
  store_prim("-", SUB);
  store_prim("and", AND);
  store_prim("or", OR);
  store_prim("xor", XOR);
  store_prim("not", NOT);
  store_prim("lshift", LSHIFT);
  store_prim("rshift", RSHIFT);
  store_prim("*", MULT);
  store_prim("/", DIV);
  store_prim("0=", EQ_ZERO);
  store_prim("=", EQ);
  store_prim("lit", LIT);
  store_prim(">r", RPUSH);
  store_prim("r>", RPOP);
  store_prim("@", FETCH);
  store_prim("!", STORE);
  store_prim(";", EXIT);  make_immediate();
  store_prim("exit", EXIT);
  store_prim(":", DEF); 
  store_prim("create", CREATE); 
  store_prim("variable", VARIABLE); 
  store_prim("<", LESS_THAN);
  store_prim(">", GREATER_THAN);
  store_prim("if", IF);  make_immediate();
  store_prim("else", ELSE);  make_immediate();
  store_prim("then", THEN);  make_immediate();
  store_prim("begin", BEGIN);  make_immediate();
  store_prim("until", UNTIL);  make_immediate();
  store_prim("while", WHILE);  make_immediate();
  store_prim("repeat", REPEAT);  make_immediate();
  store_prim("do", DO);  make_immediate();
  store_prim("leave", LEAVE);  make_immediate();
  store_prim("[']", COMPILE_FIND_WORD); make_immediate();
  store_prim("'", FIND_WORD);
  store_prim("loop", LOOP);  make_immediate();
  store_prim("i", I); make_immediate();
  store_prim("j", J); make_immediate();
  store_prim("(", COMMENT); make_immediate();
  store_prim(",\"", COMMA_STRING); make_immediate();

#endif
}




/* 
   The main exec loop. Unless a top level primitive, keep on executing
   until we hit an 'exit'.
*/
uforth_stat exec(CELL wd_idx, bool toplevelprim) {
  register CELL r1, r2;
  char b;

  while(1) {
    switch (dict->d[wd_idx++]) {
    case ABORT:
      uforth_abort();
      return E_ABORT;
    case COMMENT:
      while (uforth_next_word()) {
	if (iram.currword[iram.currwordlen-1] == ')') break;
      }
      break;
    case COMMA_STRING:
      rpush(dict->here);
      dict_append(0);	/* count */
      r1 = 0;
      b = next_char();		/* eat space */
      while (b != 0 && b!= '"') {
	r2 = 0;
	b = next_char();
	if (b == 0 || b == '"') break;
	r2 |= BYTEPACK_FIRST(b);
	++r1;
	b = next_char();
	if (b != 0 && b != '"') {
	  ++r1;
	  r2 |= BYTEPACK_SECOND(b);
	}
	dict_append(r2);
      } 
      dict_write(rpop(),r1);
      break;
    case DO:
      dict_append(SWAP);
      dict_append(RPUSH);	/* rpush limit */
      dict_append(RPUSH);	/* rpush I */
      dpush(dict->here);		/* used by LOOP to resolve */
      ++iram.donest;
      break;
    case I:
      dict_append(LIT);
      dict_append(0);
      dict_append(RPICK);
      break;
    case J:
      dict_append(LIT);
      dict_append(2);
      dict_append(RPICK);
      break;
    case LEAVE:
      dict_append(LIT);
      iram.leave_loop = dict->here;
      dict_append(0);
      dict_append(JMP);
      break;
    case LOOP:
      dict_append(RPOP);	/* rpop I */
      dict_append(LIT);
      dict_append(1);
      dict_append(ADD);		/* I+1*/
      dict_append(DUP);
      dict_append(RPUSH);	/* rpush I */
      dict_append(LIT);
      dict_append(1);
      dict_append(RPICK);	/* copy limit */
      dict_append(SWAP);
      dict_append(SUB);		/* limit - I */
      dict_append(EQ_ZERO);
      dict_append(LIT);
      if (iram.leave_loop) {
	dict_write(iram.leave_loop,dict->here + 2); /* leave -> (here+2) */
	iram.leave_loop = 0;
      }
      dict_append(dpop());	/* do address */
      dict_append(JMP_IF_ZERO);
      dict_append(RPOP);	/* ... where leave goes */
      dict_append(RPOP);
      dict_append(DROP);
      dict_append(DROP);
      --iram.donest;
      break;
    case BEGIN:
      dpush(dict->here);
      break;
    case WHILE:
      dict_append(LIT);
      dpush(dict->here);
      dict_append(0);
      dict_append(JMP_IF_ZERO);
      break;
    case REPEAT:
      dict_append(LIT);
      r1 = dpop(); 
      r2 = dpop();
      dict_append(r2);
      dict_append(JMP);
      dict_write(r1,dict->here); /* fix WHILE JMP_IF_ZERO */
      break;
    case UNTIL:
      dict_append(LIT);
      dict_append(dpop());
      dict_append(JMP_IF_ZERO);
      break;
    case IF:
      dict_append(LIT);
      dpush(dict->here);
      dict_append(0);
      dict_append(JMP_IF_ZERO);
      ++iram.ifnest;
      break;
    case ELSE:
      r1 = dpop();
      dict_write(r1, dict->here+3);
      dict_append(LIT);
      dpush(dict->here);
      dict_append(0);
      dict_append(JMP);
      break;
    case THEN:
      r1 = dpop();
      dict_write(r1, dict->here);
      --iram.ifnest;
      break;
    case CALLC:
      r1 = c_handle();
      if (r1 != OK) return r1;
      break;
    case CREATE:
      make_word(uforth_next_word(),iram.currwordlen);
      dict_append(LIT);
      r1 = dict->here + 2;	/*  next cell after EXIT */
      dict_append(r1);
      dict_append(EXIT);
      break;
    case VARIABLE:
      make_word(uforth_next_word(),iram.currwordlen);
      dict_append(LIT);
      dict_append(-1*(++iram.varidx));
      dict_append(EXIT);
      break;
    case DEF:
      iram.compiling = 1;
      return OK;
    case IMMEDIATE:
      make_immediate();
      break;
    case JMP:
      wd_idx = dpop();
      break;
    case JMP_IF_ZERO:
      r1 = dpop(); r2 = dpop();
      if (r2 == 0) wd_idx = r1;
      break;
    case COMMA:
      dict_append(dpop());
      break;
    case FIND_WORD:
      r1 = find_word(uforth_next_word(), 0, &b);
      if (b) {
	dpush(dict->d[r1]);
      } else {
	dpush(r1);
      }
      break;
    case COMPILE_FIND_WORD:
      dict_append(LIT);
      r1 = find_word(uforth_next_word(), 0, &b);
      if (b) {
	dict_append(dict->d[r1]);
      } else {
	dict_append(r1);
      }
      dict_append(COMMA);
      break;
    case HERE:
      dpush(dict->here);
      break;
    case LIT :  
      dpush(dict->d[wd_idx++]);
      break;
    case LESS_THAN_ZERO:
      dtop() = dtop() < 0;
      break;
    case ADD :
      r1 = dpop(); r2 = dtop(); 
      dtop() = r1+r2;
      break;
    case SUB :
      r1 = dpop(); r2 = dtop(); 
      dtop() = r2-r1;
      break;
    case AND:
      r1 = dpop(); r2 = dtop();
      dtop() = r1&r2;
      break;
    case LSHIFT:
      r1 = dpop(); r2 = dtop();
      dtop() = r1<<r2;
      break;
    case RSHIFT:
      r1 = dpop(); r2 = dtop(); 
      dtop() = r1>>r2;
      break;
    case OR:
      r1 = dpop(); r2 = dtop(); 
      dtop() = r1|r2;
      break;
    case XOR:
      r1 = dpop(); r2 = dtop(); 
      dtop() = r1^r2;
      break;
    case NOT:
      dtop() = ~dtop();
      break;
    case MULT :
      r1 = dpop(); r2 = dtop(); 
      dtop() = r1*r2;
      break;
    case DIV :
      r1 = dpop(); r2 = dtop(); 
      dtop() = r2/r1;
      break;
    case RPICK:
      r1 = dpop();
      r2 = iram.rs[iram.ridx - r1];
      dpush(r2);
      break;
    case PICK:
      r1 = dpop();
      r2 = iram.ds[iram.didx - r1];
      dpush(r2);
      break;
    case SWAP:
      r1 = iram.ds[iram.didx];
      iram.ds[iram.didx] = iram.ds[iram.didx-1];
      iram.ds[iram.didx-1] = r1;
      break;
    case DUP:
      r1 = dtop();
      dpush(r1);
      break;
    case DROP:
      if (iram.didx > -1) --iram.didx; else return E_STACK_UNDERFLOW;
      break;
    case EQ_ZERO:
      dtop() = (dtop() == 0);
      break;
    case RPUSH:
      rpush(dpop());
      break;
    case RPOP:
      dpush(rpop());
      break;
    case FETCH:
      r1 = dtop();
      dtop() = (r1 < 0) ? uram[abs(r1)] : dict->d[r1];
      break;
    case STORE:
      r1 = dpop();
      r2 = dpop();
      if (r1 < 0) 
	uram[abs(r1)] = r2;
      else
	dict_write(r1,r2);
      break;
    case EXEC:
      r1 = dpop();
      rpush(wd_idx);
      wd_idx = r1; if (wd_idx < 0) toplevelprim = 1;
      break;
    case EXIT:
      if (iram.ridx < 0) return OK;
      wd_idx = rpop();
      break;
    case LESS_THAN:
      r1 = dpop();
      dtop() = dtop() < r1;
      break;
    case GREATER_THAN:
      r1 = dpop();
      dtop() = dtop() > r1;
      break;
    case EQ:
      dtop() = dtop() == dtop();
      break;
    default:
      /* Execute user word by calling until we reach primitives */
      rpush(wd_idx);
      wd_idx = dict->d[wd_idx-1]; /* wd_idx-1 is current word */
      break;
    }
    if (toplevelprim) return OK;
  } /* while(1) */
}

CELL find_word(char* s, bool *immediate, bool *primitive) {
  uint8_t slen = iram.currwordlen;
  CELL fidx = dict->last_word_idx;
  CELL prev = fidx;
  uint8_t wlen;

  while (fidx > -1) {
    prev = dict->d[fidx++];
    wlen = dict->d[fidx++]; 
    if (immediate) *immediate = (wlen & IMMEDIATE_BIT) ? 1 : 0;
    if (primitive) *primitive = (wlen & PRIM_BIT) ? 1 : 0;
    wlen &= WORD_LEN_BITS;
    if (wlen == slen && strncmp(s,(char*)(dict->d+fidx),wlen) == 0) {
      fidx += (wlen / BYTES_PER_CELL) + (wlen % BYTES_PER_CELL);
      return fidx;
    }
    fidx = prev;
  }
  return -1;
}

uforth_stat uforth_interpret(char *str) {
  uforth_stat stat;
  char *word;
  CELL wd_idx;
  bool immediate = 0;
  bool primitive = 0;

  iram.inbufptr = str;
  while(*(word = uforth_next_word()) != 0) {
    wd_idx = find_word(word,&immediate,&primitive);
    switch (iram.compiling) {
    case 0:			/* interpret mode */
      if (wd_idx == -1) {	/* number or trash */
	CELL num = atoi(word);
	if (num == 0 && word[0] != '0') {
	  uforth_abort();
	  return E_NOT_A_WORD;
	}
	dpush(num);
      } else {
	stat = exec(wd_idx,primitive);
	if (stat != OK) {
	  uforth_abort();
	  return stat;
	}
      }
      break;
    case 1:			/* colon def started */
      make_word(word,iram.currwordlen);
      iram.compiling = 2;
      break;
    case 2:			/* in the middle of a colon def */
      if (wd_idx == -1) {	/* number or trash */
	CELL num = atoi(word);
	if (num == 0 && word[0] != '0') {
	  uforth_abort();
	  return E_NOT_A_WORD;
	}
	dict_append(LIT);
	dict_append(num);
      }	else if (word[0] == ';') { /* exit from a colon def */
	iram.compiling = 0;
	if (iram.ifnest > 0) {
	  uforth_abort();
	  return E_MISSING_THEN;
	}
	if (iram.donest > 0) {
	  uforth_abort();
	  return E_MISSING_LOOP;
	}
	dict_append(EXIT);
      } else if (immediate) {	/* run immediate word */
	stat = exec(wd_idx,primitive);
	if (stat != OK) {
	  uforth_abort();
	  return stat;
	}
      } else {			/* just compile word */
	if (primitive) {
	  dict_append(dict->d[wd_idx]);
	} else {
	  dict_append(wd_idx);
	}
      }
      break;
    }
  }
  return OK;
}

